@@ -24,6 +24,7 @@
 #include "pango-break.h"
 #include "pango-script-private.h"
 #include "pango-emoji-private.h"
+#include "pango-attributes-private.h"
 #include "pango-break-table.h"
 #include "pango-impl-utils.h"
 #include <string.h>
@@ -398,14 +399,13 @@ pango_default_break (const gchar   *text,
                             wc == 0x6DD ||
                             wc == 0x70F ||
                             wc == 0x8E2 ||
-                            wc == 0xD4E ||
                             wc == 0x110BD ||
-                            (wc >= 0x111C2 && wc <= 0x111C3)))
+                            wc == 0x110CD))
               {
                 GB_type = GB_Prepend;
                 break;
               }
-	    /* fall through */
+            G_GNUC_FALLTHROUGH;
 	  case G_UNICODE_CONTROL:
 	  case G_UNICODE_LINE_SEPARATOR:
 	  case G_UNICODE_PARAGRAPH_SEPARATOR:
@@ -421,11 +421,15 @@ pango_default_break (const gchar   *text,
 		GB_type = GB_ControlCRLF;
 		break;
 	      }
-	    /* fall through */
+            G_GNUC_FALLTHROUGH;
 
 	  case G_UNICODE_OTHER_LETTER:
 	    if (makes_hangul_syllable)
 	      GB_type = GB_InHangulSyllable;
+
+	    if (_pango_is_Consonant_Preceding_Repha (wc) ||
+		_pango_is_Consonant_Prefixed (wc))
+	      GB_type = GB_Prepend;
 	    break;
 
 	  case G_UNICODE_MODIFIER_LETTER:
@@ -570,7 +574,7 @@ pango_default_break (const gchar   *text,
 		    WB_type = WB_ExtendFormat; /* Other_Grapheme_Extend */
 		  break;
 		case 0x05:
-		  if (wc == 0x05F3)
+		  if (wc == 0x058A)
 		    WB_type = WB_ALetter; /* ALetter exceptions */
 		  break;
 		}
@@ -594,7 +598,7 @@ pango_default_break (const gchar   *text,
 		case G_UNICODE_CONTROL:
 		  if (wc != 0x000D && wc != 0x000A && wc != 0x000B && wc != 0x000C && wc != 0x0085)
 		    break;
-		  /* fall through */
+                  G_GNUC_FALLTHROUGH;
 		case G_UNICODE_LINE_SEPARATOR:
 		case G_UNICODE_PARAGRAPH_SEPARATOR:
 		  WB_type = WB_NewlineCRLF; /* CR, LF, Newline */
@@ -617,12 +621,16 @@ pango_default_break (const gchar   *text,
 		    WB_type = WB_MidNumLet; /* MidNumLet */
 		  break;
 		case G_UNICODE_OTHER_PUNCTUATION:
-		  if (wc == 0x0027 || wc == 0x002e || wc == 0x2024 ||
+		  if ((wc >= 0x055a && wc <= 0x055c) ||
+		      wc == 0x055e || wc == 0x05f3)
+		    WB_type = WB_ALetter; /* ALetter */
+		  else if (wc == 0x0027 || wc == 0x002e || wc == 0x2024 ||
 		      wc == 0xfe52 || wc == 0xff07 || wc == 0xff0e)
 		    WB_type = WB_MidNumLet; /* MidNumLet */
-		  else if (wc == 0x00b7 || wc == 0x05f4 || wc == 0x2027 || wc == 0x003a || wc == 0x0387 ||
+		  else if (wc == 0x00b7 || wc == 0x05f4 || wc == 0x2027 ||
+			   wc == 0x003a || wc == 0x0387 || wc == 0x055f ||
 			   wc == 0xfe13 || wc == 0xfe55 || wc == 0xff1a)
-		    WB_type = WB_MidLetter; /* WB_MidLetter */
+		    WB_type = WB_MidLetter; /* MidLetter */
 		  else if (wc == 0x066c ||
 			   wc == 0xfe50 || wc == 0xfe54 || wc == 0xff0c || wc == 0xff1b)
 		    WB_type = WB_MidNum; /* MidNum */
@@ -858,11 +866,13 @@ pango_default_break (const gchar   *text,
 
 	    if (SB_type == SB_Other)
 	      {
-		if (g_unichar_islower(wc))
+                if (type == G_UNICODE_LOWERCASE_LETTER)
 		  SB_type = SB_Lower;
-		else if (g_unichar_isupper(wc))
+                else if (type == G_UNICODE_UPPERCASE_LETTER)
 		  SB_type = SB_Upper;
-		else if (g_unichar_isalpha(wc))
+                else if (type == G_UNICODE_TITLECASE_LETTER ||
+                         type == G_UNICODE_MODIFIER_LETTER ||
+                         type == G_UNICODE_OTHER_LETTER)
 		  SB_type = SB_OLetter;
 
 		if (type == G_UNICODE_OPEN_PUNCTUATION ||
@@ -1072,10 +1082,12 @@ pango_default_break (const gchar   *text,
 	  if ((prev_break_type == G_UNICODE_BREAK_ALPHABETIC ||
 	       prev_break_type == G_UNICODE_BREAK_HEBREW_LETTER ||
 	       prev_break_type == G_UNICODE_BREAK_NUMERIC) &&
-	      break_type == G_UNICODE_BREAK_OPEN_PUNCTUATION)
+	      break_type == G_UNICODE_BREAK_OPEN_PUNCTUATION &&
+	      !_pango_is_EastAsianWide (wc))
 	    break_op = BREAK_PROHIBITED;
 
 	  if (prev_break_type == G_UNICODE_BREAK_CLOSE_PARANTHESIS &&
+	      !_pango_is_EastAsianWide (prev_wc)&&
 	      (break_type == G_UNICODE_BREAK_ALPHABETIC ||
 	       break_type == G_UNICODE_BREAK_HEBREW_LETTER ||
 	       break_type == G_UNICODE_BREAK_NUMERIC))
@@ -1217,25 +1229,7 @@ pango_default_break (const gchar   *text,
 
 	  /* Rule LB22 */
 	  if (break_type == G_UNICODE_BREAK_INSEPARABLE)
-	    {
-	      if (prev_break_type == G_UNICODE_BREAK_ALPHABETIC ||
-		  prev_break_type == G_UNICODE_BREAK_HEBREW_LETTER)
-		break_op = BREAK_PROHIBITED;
-
-	      if (prev_break_type == G_UNICODE_BREAK_EXCLAMATION)
-		break_op = BREAK_PROHIBITED;
-
-	      if (prev_break_type == G_UNICODE_BREAK_IDEOGRAPHIC ||
-		  prev_break_type == G_UNICODE_BREAK_EMOJI_BASE ||
-		  prev_break_type == G_UNICODE_BREAK_EMOJI_MODIFIER)
-		break_op = BREAK_PROHIBITED;
-
-	      if (prev_break_type == G_UNICODE_BREAK_INSEPARABLE)
-		break_op = BREAK_PROHIBITED;
-
-	      if (prev_break_type == G_UNICODE_BREAK_NUMERIC)
-		break_op = BREAK_PROHIBITED;
-	    }
+	    break_op = BREAK_PROHIBITED;
 
 	  if (break_type == G_UNICODE_BREAK_AFTER ||
 	      break_type == G_UNICODE_BREAK_HYPHEN ||
@@ -1603,7 +1597,7 @@ tailor_break (const char    *text,
 
   res = break_script (text, length, analysis, attrs, attrs_len);
 
-  if (item_offset >= 0)
+  if (item_offset >= 0 && analysis->extra_attrs)
     res |= break_attrs (text, length, analysis->extra_attrs, item_offset, attrs, attrs_len);
 
   return res;
@@ -1912,31 +1906,36 @@ break_attrs (const char   *text,
              PangoLogAttr *log_attrs,
              int           log_attrs_len)
 {
-  PangoAttrList *list;
-  PangoAttrIterator *iter;
+  PangoAttrList list;
+  PangoAttrIterator iter;
   GSList *l;
 
-  list = pango_attr_list_new ();
+  _pango_attr_list_init (&list);
   for (l = attributes; l; l = l->next)
     {
       PangoAttribute *attr = l->data;
 
       if (attr->klass->type == PANGO_ATTR_ALLOW_BREAKS)
-        pango_attr_list_insert (list, pango_attribute_copy (attr));
+        pango_attr_list_insert (&list, pango_attribute_copy (attr));
     }
 
-  iter = pango_attr_list_get_iterator (list);
+  if (!_pango_attr_list_has_attributes (&list))
+    {
+      _pango_attr_list_destroy (&list);
+      return FALSE;
+    }
+
+  _pango_attr_list_get_iterator (&list, &iter);
   do {
-    PangoAttribute *attr;
+    const PangoAttribute *attr = pango_attr_iterator_get (&iter, PANGO_ATTR_ALLOW_BREAKS);
 
-    attr = pango_attr_iterator_get (iter, PANGO_ATTR_ALLOW_BREAKS);
     if (attr && ((PangoAttrInt*)attr)->value == 0)
       {
         int start, end;
         int start_pos, end_pos;
         int pos;
 
-        pango_attr_iterator_range (iter, &start, &end);
+        pango_attr_iterator_range (&iter, &start, &end);
         if (start < offset)
           start_pos = 0;
         else
@@ -1953,10 +1952,10 @@ break_attrs (const char   *text,
             log_attrs[pos].is_char_break = FALSE;
           }
       }
-  } while (pango_attr_iterator_next (iter));
+  } while (pango_attr_iterator_next (&iter));
 
-  pango_attr_iterator_destroy (iter);
-  pango_attr_list_unref (list);
+  _pango_attr_iterator_destroy (&iter);
+  _pango_attr_list_destroy (&list);
 
   return TRUE;
 }
