@@ -238,12 +238,8 @@ test_file (const gchar *filename, GString *string)
   PangoWrapMode wrap = PANGO_WRAP_WORD;
   PangoFontDescription *desc;
 
-  if (!g_file_get_contents (filename, &contents, &length, &error))
-    {
-      fprintf (stderr, "%s\n", error->message);
-      g_error_free (error);
-      return;
-    }
+  g_file_get_contents (filename, &contents, &length, &error);
+  g_assert_no_error (error);
 
   p = strchr (contents, '\n');
   g_assert (p);
@@ -308,6 +304,19 @@ test_layout (gconstpointer d)
   GString *dump;
   gchar *diff;
 
+  const char *old_locale = setlocale (LC_ALL, NULL);
+  setlocale (LC_ALL, "en_US.utf8");
+  if (strstr (setlocale (LC_ALL, NULL), "en_US") == NULL)
+    {
+      char *msg = g_strdup_printf ("Locale en_US.UTF-8 not available, skipping layout %s", filename);
+      g_test_skip (msg);
+      g_free (msg);
+      return;
+    }
+
+  if (context == NULL)
+    context = pango_font_map_create_context (pango_cairo_font_map_get_default ());
+
   expected_file = get_expected_filename (filename);
 
   dump = g_string_sized_new (0);
@@ -317,10 +326,21 @@ test_layout (gconstpointer d)
   diff = diff_with_file (expected_file, dump->str, dump->len, &error);
   g_assert_no_error (error);
 
+  setlocale (LC_ALL, old_locale);
+
   if (diff && diff[0])
     {
-      g_printerr ("Contents don't match expected contents:\n%s", diff);
+      char **lines = g_strsplit (diff, "\n", -1);
+      const char *line;
+      int i = 0;
+
+      g_test_message ("Contents don't match expected contents");
+
+      for (line = lines[0]; line != NULL; line = lines[++i])
+        g_test_message ("%s", line);
+
       g_test_fail ();
+      g_strfreev (lines);
       g_free (diff);
     }
 
@@ -336,13 +356,8 @@ main (int argc, char *argv[])
   const gchar *name;
   gchar *path;
 
-  g_setenv ("LC_ALL", "en_US.UTF-8", TRUE);
-  setlocale (LC_ALL, "");
-
   g_test_init (&argc, &argv, NULL);
 
-  context = pango_font_map_create_context (pango_cairo_font_map_get_default ());
-
   /* allow to easily generate expected output for new test cases */
   if (argc > 1)
     {
@@ -350,7 +365,7 @@ main (int argc, char *argv[])
 
       string = g_string_sized_new (0);
       test_file (argv[1], string);
-      g_print ("%s", string->str);
+      g_test_message ("%s", string->str);
 
       return 0;
     }
