@@ -618,6 +618,218 @@ test_list_update (void)
   pango_attr_list_unref (list);
 }
 
+static void
+test_list_equal (void)
+{
+  PangoAttrList *list1, *list2;
+  PangoAttribute *attr;
+
+  list1 = pango_attr_list_new ();
+  list2 = pango_attr_list_new ();
+
+  g_assert_true (pango_attr_list_equal (NULL, NULL));
+  g_assert_false (pango_attr_list_equal (list1, NULL));
+  g_assert_false (pango_attr_list_equal (NULL, list1));
+  g_assert_true (pango_attr_list_equal (list1, list1));
+  g_assert_true (pango_attr_list_equal (list1, list2));
+
+  attr = pango_attr_size_new (10 * PANGO_SCALE);
+  attr->start_index = 0;
+  attr->end_index = 7;
+  pango_attr_list_insert (list1, pango_attribute_copy (attr));
+  pango_attr_list_insert (list2, pango_attribute_copy (attr));
+  pango_attribute_destroy (attr);
+
+  g_assert_true (pango_attr_list_equal (list1, list2));
+
+  attr = pango_attr_stretch_new (PANGO_STRETCH_CONDENSED);
+  attr->start_index = 0;
+  attr->end_index = 1;
+  pango_attr_list_insert (list1, pango_attribute_copy (attr));
+  g_assert_true (!pango_attr_list_equal (list1, list2));
+
+  pango_attr_list_insert (list2, pango_attribute_copy (attr));
+  g_assert_true (pango_attr_list_equal (list1, list2));
+  pango_attribute_destroy (attr);
+
+  attr = pango_attr_size_new (30 * PANGO_SCALE);
+  /* Same range as the first attribute */
+  attr->start_index = 0;
+  attr->end_index = 7;
+  pango_attr_list_insert (list2, pango_attribute_copy (attr));
+  g_assert_true (!pango_attr_list_equal (list1, list2));
+  pango_attr_list_insert (list1, pango_attribute_copy (attr));
+  g_assert_true (pango_attr_list_equal (list1, list2));
+  pango_attribute_destroy (attr);
+
+  pango_attr_list_unref (list1);
+  pango_attr_list_unref (list2);
+
+
+  /* Same range but different order */
+  {
+    PangoAttrList *list1, *list2;
+    PangoAttribute *attr1, *attr2;
+
+    list1 = pango_attr_list_new ();
+    list2 = pango_attr_list_new ();
+
+    attr1 = pango_attr_size_new (10 * PANGO_SCALE);
+    attr2 = pango_attr_stretch_new (PANGO_STRETCH_CONDENSED);
+
+    pango_attr_list_insert (list1, pango_attribute_copy (attr1));
+    pango_attr_list_insert (list1, pango_attribute_copy (attr2));
+
+    pango_attr_list_insert (list2, pango_attribute_copy (attr2));
+    pango_attr_list_insert (list2, pango_attribute_copy (attr1));
+
+    pango_attribute_destroy (attr1);
+    pango_attribute_destroy (attr2);
+
+    g_assert_true (pango_attr_list_equal (list1, list2));
+    g_assert_true (pango_attr_list_equal (list2, list1));
+
+    pango_attr_list_unref (list1);
+    pango_attr_list_unref (list2);
+  }
+}
+
+static void
+test_insert (void)
+{
+  PangoAttrList *list;
+  PangoAttribute *attr;
+
+  list = pango_attr_list_new ();
+  attr = pango_attr_size_new (10 * PANGO_SCALE);
+  attr->start_index = 10;
+  attr->end_index = 11;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_rise_new (100);
+  attr->start_index = 0;
+  attr->end_index = 200;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_family_new ("Times");
+  attr->start_index = 5;
+  attr->end_index = 15;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_fallback_new (FALSE);
+  attr->start_index = 11;
+  attr->end_index = 100;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_stretch_new (PANGO_STRETCH_CONDENSED);
+  attr->start_index = 30;
+  attr->end_index = 60;
+  pango_attr_list_insert (list, attr);
+
+  assert_attr_list (list, "[0,200]rise=100\n"
+                          "[5,15]family=Times\n"
+                          "[10,11]size=10240\n"
+                          "[11,100]fallback=0\n"
+                          "[30,60]stretch=2\n");
+
+  attr = pango_attr_family_new ("Times");
+  attr->start_index = 10;
+  attr->end_index = 25;
+  pango_attr_list_change (list, attr);
+
+  assert_attr_list (list, "[0,200]rise=100\n"
+                          "[5,25]family=Times\n"
+                          "[10,11]size=10240\n"
+                          "[11,100]fallback=0\n"
+                          "[30,60]stretch=2\n");
+
+  attr = pango_attr_family_new ("Futura");
+  attr->start_index = 11;
+  attr->end_index = 25;
+  pango_attr_list_insert (list, attr);
+
+  assert_attr_list (list, "[0,200]rise=100\n"
+                          "[5,25]family=Times\n"
+                          "[10,11]size=10240\n"
+                          "[11,100]fallback=0\n"
+                          "[11,25]family=Futura\n"
+                          "[30,60]stretch=2\n");
+
+  pango_attr_list_unref (list);
+}
+
+static gboolean
+attr_list_merge_filter (PangoAttribute *attribute,
+                        gpointer        list)
+{
+  pango_attr_list_change (list, pango_attribute_copy (attribute));
+  return FALSE;
+}
+
+/* test something that gtk does */
+static void
+test_merge (void)
+{
+  PangoAttrList *list;
+  PangoAttrList *list2;
+  PangoAttribute *attr;
+
+  list = pango_attr_list_new ();
+  attr = pango_attr_size_new (10 * PANGO_SCALE);
+  attr->start_index = 10;
+  attr->end_index = 11;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_rise_new (100);
+  attr->start_index = 0;
+  attr->end_index = 200;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_family_new ("Times");
+  attr->start_index = 5;
+  attr->end_index = 15;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_fallback_new (FALSE);
+  attr->start_index = 11;
+  attr->end_index = 100;
+  pango_attr_list_insert (list, attr);
+  attr = pango_attr_stretch_new (PANGO_STRETCH_CONDENSED);
+  attr->start_index = 30;
+  attr->end_index = 60;
+  pango_attr_list_insert (list, attr);
+
+  assert_attr_list (list, "[0,200]rise=100\n"
+                          "[5,15]family=Times\n"
+                          "[10,11]size=10240\n"
+                          "[11,100]fallback=0\n"
+                          "[30,60]stretch=2\n");
+
+  list2 = pango_attr_list_new ();
+  attr = pango_attr_size_new (10 * PANGO_SCALE);
+  attr->start_index = 11;
+  attr->end_index = 13;
+  pango_attr_list_insert (list2, attr);
+  attr = pango_attr_size_new (11 * PANGO_SCALE);
+  attr->start_index = 13;
+  attr->end_index = 15;
+  pango_attr_list_insert (list2, attr);
+  attr = pango_attr_size_new (12 * PANGO_SCALE);
+  attr->start_index = 40;
+  attr->end_index = 50;
+  pango_attr_list_insert (list2, attr);
+
+  assert_attr_list (list2, "[11,13]size=10240\n"
+                           "[13,15]size=11264\n"
+                           "[40,50]size=12288\n");
+
+  pango_attr_list_filter (list2, attr_list_merge_filter, list);
+
+  assert_attr_list (list, "[0,200]rise=100\n"
+                          "[5,15]family=Times\n"
+                          "[10,13]size=10240\n"
+                          "[11,100]fallback=0\n"
+                          "[13,15]size=11264\n"
+                          "[30,60]stretch=2\n"
+                          "[40,50]size=12288\n");
+
+  pango_attr_list_unref (list);
+  pango_attr_list_unref (list2);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -630,6 +842,9 @@ main (int argc, char *argv[])
   g_test_add_func ("/attributes/list/splice", test_list_splice);
   g_test_add_func ("/attributes/list/filter", test_list_filter);
   g_test_add_func ("/attributes/list/update", test_list_update);
+  g_test_add_func ("/attributes/list/equal", test_list_equal);
+  g_test_add_func ("/attributes/list/insert", test_insert);
+  g_test_add_func ("/attributes/list/merge", test_merge);
   g_test_add_func ("/attributes/iter/basic", test_iter);
   g_test_add_func ("/attributes/iter/get", test_iter_get);
   g_test_add_func ("/attributes/iter/get_font", test_iter_get_font);
